from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

def one_hot(labels: torch.Tensor,
            num_classes: int,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            eps: Optional[float] = 1e-6) -> torch.Tensor:
    """
        Convert an integer label tensor to a one hot tensor.

        Args:
            labels (torch.Tensor) : tensor with labels of shape (n_positve+n_negative) indicating the class with an index [0, 1].
            num_classes (int): number of classes in labels.
            device (Optional[torch.device]): the desired device of the returned tensor.
            dtype (Optional[torch.dtype]): the desired data type of the returned tensor.
            eps (Optional[float]): small value to avoid division by zero.

        Returns:
            torch.Tensor: the labels in one hot tensor [1,  -> [[0, 1]
                                                        1,       [0, 1]
                                                        1,       [0, 1]
                                                        0,       [1, 0]
                                                        0,       [1, 0]
                                                        1,]      [0, 1]]
    """
    # Data validation
    if not torch.is_tensor(labels):
        raise TypeError("Input labels type is not a torch.Tensor. Got {}".format(type(labels)))
    if not labels.dtype == torch.int64:
        raise ValueError("labels must be of the same dtype torch.int64. Got: {}".format(labels.dtype))
    if num_classes < 1:
        raise ValueError("The number of classes must be bigger than one." " Got: {}".format(num_classes))
    
    # Compute the one hot tensor
    shape = labels.shape
    one_hot = torch.zeros(shape[0], num_classes, *shape[1:], device=device, dtype=dtype)
    return one_hot.scatter_(1, labels.unsqueeze(1), 1.0) + eps

def focal_loss(input: torch.Tensor,
               target: torch.Tensor,
               alpha: float,
               gamma: float = 2.0,
               eps: float = 1e-8) -> torch.Tensor:
    """
        Function that computes Focal loss.

        Args:
            input (torch.Tensor): the input tensor with the logits.
            target (torch.Tensor): the target tensor with the labels indices.
            alpha (float): Weighting factor in range (0,1) to balance positive vs negative samples.
            gamma (float): Focusing parameter for modulating factor (1 - p_t)^gamma.
            eps (float): Scalar to enforce numerical stability.

        Returns:
            torch.Tensor: the computed loss for each proposal.
    """
    if not torch.is_tensor(input):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(type(input)))

    if not len(input.shape) >= 2:
        raise ValueError("Invalid input shape, we expect BxCx*. Got: {}".format(input.shape))

    if input.size(0) != target.size(0):
        raise ValueError('Expected input batch_size ({}) to match target batch_size ({}).'.format(
            input.size(0), target.size(0)))

    # Get the number of proposals
    n = input.size(0)
    out_size = (n, ) + input.size()[2:]
    if target.size()[1:] != input.size()[2:]:
        raise ValueError('Expected target size {}, got {}'.format(out_size, target.size()))

    if not input.device == target.device:
        raise ValueError("input and target must be in the same device. Got: {} and {}".format(
            input.device, target.device))

    # Compute softmax over the classes axis
    input_soft = F.softmax(input, dim=1) + eps

    # Create the labels one hot tensor
    target_one_hot = one_hot(target, num_classes=input.shape[1], device=input.device, dtype=input.dtype)

    # Compute the actual focal loss
    weight = torch.pow(-input_soft + 1., gamma)
    focal = -alpha * weight * torch.log(input_soft)

    # Compute the loss summing over all classes since the one hot tensor will discard the negative class
    loss = torch.sum(target_one_hot * focal, dim=1)
    return loss

class FocalLoss(nn.Module):
    def __init__(self, alpha: float, gamma: float = 2.0) -> None:
        """
            Focal loss function initialization.

            Args:
                alpha (float): Weighting factor in range (0,1) to balance positive vs negative samples.
                gamma (float): Focusing parameter for modulating factor (1 - p_t)^gamma.
        """
        super(FocalLoss, self).__init__()
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.eps: float = 1e-6

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
            Forward pass of the focal loss.

            Args:
                input (torch.Tensor): the input tensor with the logits.
                target (torch.Tensor): the target tensor with the labels indices.

            Returns:
                torch.Tensor: the computed loss for each proposal.
        """
        return focal_loss(input, target, self.alpha, self.gamma, self.eps)


from torch.utils.data import Dataset
from torchvision.transforms import ToTensor
from scipy.interpolate import InterpolatedUnivariateSpline

import cv2
import json
import logging
import os
import random

import numpy as np

class LaneDataset(Dataset):
    def __init__(self, config:str, split:str) -> None:
        """
            Initialize the LaneDataset object.

            Args:
                config: The path to the configuration file.
                split: The dataset split to load, e.g. train, val, test.
        """
        self.__general_config = config
        self.__dataset_config = config['dataset'][split]
        # Dataset split to load, e.g. train, val, test
        self.__split = split
        # The root directory of the dataset
        self.__root = self.__dataset_config['root']
        # A logger object
        self.__logger = logging.getLogger(__name__)

        # Image width and height (model's input size and dataset's image size)
        self.img_w, self.img_h = self.__general_config['image_size']['width'], self.__general_config['image_size']['height']
        self.dataset_img_w, self.dataset_img_h = self.__general_config['dataset_image_size']['width'], self.__general_config['dataset_image_size']['height']

        # Load the annotation files
        jsons = os.listdir(self.__root)
        jsons = [j for j in jsons if j.endswith('.json')]
        self.__annotation_files = [os.path.join(self.__root, json) for json in jsons]

        # Verify that the root directory is specified
        if self.__root is None:
            raise Exception('Please specify the root directory')

        # Load the annotations
        self.annotations = []
        self.max_lanes = 0
        self.__load_annotations()

        # Transform the annotations to the model's target format
        self.__logger.info("Transforming annotations to the model's target format...")
        self.__y_discretizations = self.__general_config['anchor_discretization']['y']
        self.__ys = np.linspace(1, 0, self.__y_discretizations)
        self.annotations = np.array(list(map(self.__transform_annotation, self.annotations)))
        self.__logger.info("Annotations transformed.")

    def __load_annotations(self) -> None:
        """
            Load annotations from the annotation files as a dict of the path of the image and a list of lists of tuples (x,y) for the lanes.

            annotations: {
                            path: str,
                            lanes: list[list[tuple(float, float)]]
                            }
        """

        self.__logger.info('Loading TuSimple annotations...')
        max_lanes = 0
        self.annotations = []
        # Iterate over the annotation files
        for annotation_file in self.__annotation_files:
            # Opens and reads the annotation file
            with open(annotation_file, 'r') as annotation_obj:
                lines = annotation_obj.readlines()
            # Iterate over the lines in the annotation file
            for line in lines:
                # Load the JSON line data
                try:
                    data = json.loads(line)
                except:
                    print('Error loading JSON line:', line)
                # Get the lanes y coordinates
                y_samples = data['h_samples']
                # Get the lanes x coordinates
                og_lanes = data['lanes']
                # Create the lanes as a tuple of x and y coordinates
                lanes = [[(x, y) for (x, y) in zip(lane, y_samples) if x >= 0] for lane in og_lanes]
                # Get the maximum number of lanes in an image
                max_lanes = max(max_lanes, len(lanes))

                # Append the annotation to the list of annotations
                self.annotations.append({
                    'path': os.path.join(self.__root, data['raw_file']),
                    'lanes': lanes,
                })

        # Shuffle the annotations if the split is train
        if self.__split == 'train':
            random.shuffle(self.annotations)
        self.max_lanes = max_lanes
        self.__logger.info('%d annotations loaded, with a maximum of %d lanes in an image.', len(self.annotations),
                         self.max_lanes)

    def __transform_annotation(self, annotation:dict) -> dict:
        """
            Transform the annotation to the model's target format.
            
            The transformed lanes involve cleaning the lanes, sorting them by Y, removing points with the same Y, 
            remapping the lanes to a 0-1 range, and creating a new annotation with the transformed lanes and the model
            y_discretizations based on polynomial interpolations and extrapolations of the labeled lanes.

            It also involves labeling valid lanes in the score fields, setting the starting point of the lanes (based on
            the first x inside the image and its corresponding y assuming bottom approximations only can leave the image
            by one side), and its length.

            Args:
                annotation: A dictionary containing the path of the image and the lanes.

            Returns:
                A dictionary containing the path of the image and the transformed lanes.Returns:
                A dictionary containing the path of the image and the transformed lanes.
        """
        old_lanes = annotation['lanes']

        # Remove lanes with less than 2 points
        old_lanes = filter(lambda x: len(x) > 1, old_lanes)
        # Sort lane points by Y (bottom to top of the image)
        old_lanes = [sorted(lane, key=lambda x: -x[1]) for lane in old_lanes]
        # Remove points with same Y (keep first occurrence)
        old_lanes = [self.__filter_lane(lane) for lane in old_lanes]
        # Remap the lanes to the model config image size
        old_lanes = [[(x/self.dataset_img_w, y/self.dataset_img_h) for (x, y) in lane] for lane in old_lanes]
        # Create tranformed annotations placeholder
        lanes = np.ones((self.max_lanes, 2 + 1 + 1 + 1 + self.__y_discretizations),
                        dtype=np.float32) * -1e5
        
        # Lanes are invalid by default
        lanes[:, 0] = 1
        lanes[:, 1] = 0
        # Iterates over the lanes
        for lane_idx, lane in enumerate(old_lanes):
            try:
                xs_outside_image, xs_inside_image = self.__sample_lane(lane)
            except AssertionError:
                continue
            if len(xs_inside_image) == 0:
                continue
            # Joins the points outside and inside the image again
            all_xs = np.hstack((xs_outside_image, xs_inside_image))

            # Updates the lanes with the new points
            # Sets the scores to 0 for no lane and 1 for lane
            lanes[lane_idx, 0] = 0
            lanes[lane_idx, 1] = 1
            # Set y starting point bottom to top
            lanes[lane_idx, 2] = (1-self.__ys[len(xs_outside_image)]) * self.img_h
            # Set x starting point
            lanes[lane_idx, 3] = xs_inside_image[0] * self.img_w
            # Set the number of points in the lane
            lanes[lane_idx, 4] = len(xs_inside_image)
            # Set the x coordinates of the lane
            lanes[lane_idx, 5:5 + len(all_xs)] = all_xs * self.img_w

        new_annotation = {'path': annotation['path'], 'label': lanes}
        return new_annotation
    
    def __filter_lane(self, lane:list) -> list:
        """
            Remove points with the same Y coordinate, keeping only the first occurrence.

            Args:
                lane: A list of points representing a lane.

            Returns:
                filtered_lane: A list of points representing a lane without points with the same Y coordinate.
        """
        assert lane[-1][1] <= lane[0][1] # Invalid lane
        filtered_lane = []
        used = set()
        for p in lane:
            if p[1] not in used:
                filtered_lane.append(p)
                used.add(p[1])

        return filtered_lane
    
    def __sample_lane(self, points:list) -> tuple:
        """
            Sample the lane points at the anchor points using the interpolation function and extrapolate the lane with a straight line.
            It also separates the points inside and outside the image.

            Args:
                points: A list of points representing a lane.
            
            Returns:    
                xs_outside_image: The x coordinates of the lane points outside the image.
                xs_inside_image: The x coordinates of the lane points inside the image.
        """
        points = np.array(points)
        x, y = points[:, 0], points[:, 1]

        # Interpolate points inside domain
        assert len(points) > 1
        interp = InterpolatedUnivariateSpline(y[::-1], x[::-1], k=min(3, len(points) - 1))
        domain_min_y = y.min()
        domain_max_y = y.max()
        # Get anchor points inside the domain of the labeled lane
        sample_ys_inside_domain = self.__ys[(self.__ys >= domain_min_y) & (self.__ys <= domain_max_y)]
        assert len(sample_ys_inside_domain) > 0
        # Evaluates the y samples in the interpolation function to get the x samples
        interp_xs = interp(sample_ys_inside_domain)

        # Extrapolate lane with a straight line using the 2 points closest to the bottom
        two_closest_points = points[:2]
        extrap = np.polyfit(two_closest_points[:, 1], two_closest_points[:, 0], deg=1)
        extrap_ys = self.__ys[self.__ys > domain_max_y]
        extrap_xs = np.polyval(extrap, extrap_ys)
        all_xs = np.hstack((extrap_xs, interp_xs))

        # Separate between inside and outside points
        inside_mask = (all_xs >= 0) & (all_xs < 1)
        xs_inside_image = all_xs[inside_mask]
        xs_outside_image = all_xs[~inside_mask]

        return xs_outside_image, xs_inside_image

    def __getitem__(self, idx):
        """
            Get an item from the dataset.

            Args:
                idx: The index of the item to get.

            Returns:
                A tuple containing the image (matrix) and the label (tensor of lane format).
        """
        item = self.annotations[idx]
        img_org = cv2.imread(item['path'])
        img_org = cv2.resize(img_org, (self.img_w, self.img_h))
        img = ToTensor()((img_org.copy()/255.0).astype(np.float32))
        label = item['label']
        return (img, label)
    
    def __len__(self):
        return len(self.annotations)
    



from typing import List

import torch

import numpy as np

def generate_anchors(y_discretization:int, x_discretization:int, left_angles:list, right_angles:list, bottom_angles:list, fv_size:tuple, img_size:tuple) -> List[torch.Tensor]:
    """
        Generates anchors for the model based on the discretization and angles
        The anchors are in the form of a tensor with shape 
        (y_discretization*[len(left_angles)+len(right_angles)]+x_discretization*len(bottom_angles), (y_discretization+5 or feature volume height+5)) 
        where each row contains each anchor information and y_(y_discretization+5 or feature volume height+5) is the number of features for each anchor
        
        Structure of each anchor:
        [score0, score1, start_y, start_x, length, anchor_xs]
        where:
        - score0: No line score
        - score1: Line score
        - start_y: Start y coordinate (bottom is 0, top is image height or feature volume height)
        - start_x: Start x coordinate (left is 0, right is image width or feature volume width)
        - length: Length of the anchor (steps in the discretization)
        - anchor_xs: X coordinates of the anchor points (image pixel coordinates or feature volume pixel coordinates) [Bottom to top]

        Args:
            y_discretization (int): Number of steps in y direction
            x_discretization (int): Number of steps in x direction
            left_angles (list): List of left angles
            right_angles (list): List of right angles
            bottom_angles (list): List of bottom angles
            fv_size (tuple): Feature volume size
            img_size (tuple): Image size
        
        Returns:
            torch.Tensor: Anchors for all sides projected into the image
            torch.Tensor: Anchors for all sides projected into the feature volume
    """

    # Generate left anchors
    left_anchors_image, left_anchors_volume = generate_side_anchors(left_angles, y_discretization, fv_size, y_discretization, img_size, x=0.)
    # Generate right anchors
    right_anchors_image, right_anchors_volume = generate_side_anchors(right_angles, y_discretization, fv_size, y_discretization, img_size, x=1.)
    # Generate bottom anchors
    bottom_anchors_image, bottom_anchors_volume = generate_side_anchors(bottom_angles, x_discretization, fv_size, y_discretization, img_size, y=1.)

    # Concatenate anchors and cut anchors
    return torch.cat([left_anchors_image, bottom_anchors_image, right_anchors_image]), torch.cat([left_anchors_volume, bottom_anchors_volume, right_anchors_volume]) 

def generate_side_anchors(angles:list, discretization:int, fv_size:tuple, y_discretization:int, img_size:tuple, x:float=None, y:float=None) -> List[torch.Tensor]:
    """
        Generates side anchors based on predefined angles, and discretization
        The anchors are in the form of a tensor with shape (discretization*len(angles), y_(y_discretization+5 or feature volume height+5)) 
        where each row contains each anchor information and y_(y_discretization+5 or feature volume height+5) is the number of features for each anchor
        
        Structure of each anchor:
        [score0, score1, start_y, start_x, length, anchor_xs]
        where:
        - score0: No line score
        - score1: Line score
        - start_y: Start y coordinate (bottom is 0, top is image height or feature volume height)
        - start_x: Start x coordinate (left is 0, right is image width or feature volume width)
        - length: Length of the anchor (steps in the discretization)
        - anchor_xs: X coordinates of the anchor points (image pixel coordinates or feature volume pixel coordinates) [Bottom to top]

        Args:
            angles (list): List of angles
            discretization (int): Number of steps in the side direction
            fv_size (tuple): Feature volume size
            y_discretization (int): Number of steps in y direction
            img_size (tuple): Image size
            x (float): X coordinate of the side
            y (float): Y coordinate of the side

        Returns:
            torch.Tensor: Anchors for the side projected into the image
            torch.Tensor: Anchors for the side projected into the feature volume
    """

    # Check if x or y is None
    if x is None and y is not None:
        # Generate starts based on a fixed y
        starts = [(x, y) for x in np.linspace(1., 0., num=discretization)]
    elif x is not None and y is None:
        # Generate starts based on a fixed x
        starts = [(x, y) for y in np.linspace(1., 0., num=discretization)]
    else:
        # Raises an error if no side is defined
        raise Exception('Please define exactly one of `x` or `y` (not neither nor both)')

    # Calculate number of anchors since one anchor is generated for each angle and origin
    anchors_number = discretization * len(angles)

    # Initialize anchors as a tensor of anchors_number as rows and (y_discretization or feature map height + 5) as columns.
    # This represents each anchor list will have 2 scores, 1 start y, 1 start x, 1 length and y_discretization or feature map height x coordinates
    anchors_image = torch.zeros((anchors_number, 2 + 2 + 1 + y_discretization))
    anchors_feature_volume = torch.zeros((anchors_number, 2 + 2 + 1 + fv_size[1]))

    # Iterates over each start point
    for i, start in enumerate(starts):
        # Iterates over each angle for each start point
        for j, angle in enumerate(angles):
            # Calculates the index of the anchor
            k = i * len(angles) + j
            # Generates the anchors
            anchors_image[k] = generate_anchor(start, angle, y_discretization, fv_size, img_size)
            anchors_feature_volume[k] = generate_anchor(start, angle, y_discretization, fv_size, img_size, fv=True)

    return anchors_image, anchors_feature_volume

def generate_anchor(start:tuple, angle:float, y_discretization:int, fv_size:int, img_size:tuple, fv:bool=False) -> torch.Tensor:
    """
        Generates anchor based on start point and angle
        The anchors are in the form of a tensor with shape (1, y_(y_discretization+5 or feature volume height+5)) 
        where y_(y_discretization+5 or feature volume height+5) is the number of features for each anchor
        
        Structure of each anchor:
        [score0, score1, start_y, start_x, length, anchor_xs]
        where:
        - score0: No line score
        - score1: Line score
        - start_y: Start y coordinate (bottom is 0, top is image height or feature volume height)
        - start_x: Start x coordinate (left is 0, right is image width of feature volume width)
        - length: Length of the anchor (steps in the discretization)
        - anchor_xs: X coordinates of the anchor points (image pixel coordinates) [Bottom to top]

        Args:
            start (tuple): Start point (x -> left is 0, right is 1, y -> top is 1, bottom is 0)
            angle (float): Angle of the anchor
            y_discretization (int): Number of steps in y direction
            fv_size (tuple): Feature volume size
            img_size (tuple): Image size
            fv (bool): Flag to indicate if the anchor is projected into the feature volume

        Returns:
            torch.Tensor: Anchor projected into the image or feature volume
    """

    # Extract image width and height
    img_h, img_w = img_size
    # Extract feature volume data
    _, fv_h, fv_w = fv_size
    # Convert angle to radians
    angle = angle * np.pi / 180.
    # Extract start x and y from start point
    start_x, start_y = start

    # Check if fv is True
    if fv:
        # Set anchor y coordinates from 1 to 0 with feature map height steps
        anchor_ys = torch.linspace(1, 0, steps=fv_h, dtype=torch.float32)
        # Initialize anchor tensor with 2 scores, 1 start y, 1 start x, 1 length and feature volume height
        anchor = torch.zeros(2 + 2 + 1 + fv_h)
        # Assigns to third element of anchor tensor the start y in feature volume coordinates
        anchor[2] = (1 - start_y)  * fv_h
        # Assigns to fourth element of anchor tensor the start x in feature volume coordinates
        anchor[3] = start_x * fv_w
        # Gets a relative delta y based on the start coordinate for each fv height point
        delta_y = (start_y - anchor_ys) * fv_h
    else:
        # Set anchor y coordinates from 1 to 0 with n_offsets steps
        anchor_ys = torch.linspace(1, 0, steps=y_discretization, dtype=torch.float32)
        # Initialize anchor tensor with 2 scores, 1 start y, 1 start x, 1 length and discretization points
        anchor = torch.zeros(2 + 2 + 1 + y_discretization)
        # Assigns to third element of anchor tensor the start y in image coordinates
        anchor[2] = (1 - start_y)  * img_h
        # Assigns to fourth element of anchor tensor the start x in image coordinates
        anchor[3] = start_x * img_w
        # Gets a relative delta y based on the start coordinate for each discretization points
        delta_y = (start_y - anchor_ys) * img_h

    # Gets a relative delta x from the origin point for each anchor point based on the angle and delta y since -> 1/tan(angle) = delta x / delta y
    delta_x = delta_y / np.tan(angle)
    # Adds the delta x of each anchor point to the start x to get the x coordinate of each anchor point
    anchor[5:] = (anchor[3] + delta_x)

    return anchor

def get_fv_anchor_indices(anchors_fv, feature_map_channels, feature_volume_height, feature_volume_width):
        """
            Computes the anchor indices for the feature volume

            Args:
                anchors_fv (torch.Tensor): Anchors for all sides projected into the feature volume
                feature_map_channels (int): Number of feature map channels
                feature_volume_height (int): Feature volume height
                feature_volume_width (int): Feature volume width
            
            Returns:
                torch.Tensor: Z indices ((0-64)*fv_height[interleaved]*n_proposals, 1)
                torch.Tensor: Y indices ((0-23)*fv_channels*n_proposals, 1)
                torch.Tensor: X indices (ith_proposal[5:]*fv_channels*n_proposals, 1)
                torch.Tensor: Invalid mask (ith_proposal[5:]*fv_channels*n_proposals, 1)
        """
        # Get the number of anchors proposals
        n_proposals = len(anchors_fv)

        # Extract only anchor points from anchors_feature_volume tensor
        anchors_x_points = anchors_fv[:, 5:]
        # Repeat the anchors proposals for each feature map and puts them in a single dimension
        # The output of this line stores the ith anchor proposal repeated for each feature map channel collapsed in a single dimension
        # (ith_propsal[5:], fv_h) -> (ith_proposal[5:]*fv_channels, 1) and rounded to the nearest integer to get the nearest pixel index
        unclamped_anchors_x_indices = torch.repeat_interleave(anchors_x_points, feature_map_channels, dim=0).reshape(-1, 1).round().long()
        # Clamp the anchors coordinates to the feature map width
        anchors_x_indices = torch.clamp(unclamped_anchors_x_indices, 0, feature_volume_width - 1)
        # Reshape the anchors to the original shape
        unclamped_anchors_x_indices = unclamped_anchors_x_indices.reshape(n_proposals, feature_map_channels, feature_volume_height, 1)
        
        # Generate a binary mask to filter out the invalid anchor proposals
        invalid_mask = (unclamped_anchors_x_indices < 0) | (unclamped_anchors_x_indices > feature_volume_width - 1)

        # Generate y coordinates for each anchor point
        anchors_y_indices = torch.arange(0, feature_volume_height)
        # Repeat the y coordinates for each feature map and each proposal and puts them in a single dimension
        # ((0-23)*fv_channels*n_proposals, 1)
        anchors_y_indices = anchors_y_indices.repeat(feature_map_channels).repeat(n_proposals).reshape(-1, 1)

        # Generate z pixels for each anchor proposal and puts them in a single dimension
        # ((0-64)*fv_height[interleaved]*n_proposals, 1)
        anchors_z_cut_indices = torch.arange(feature_map_channels).repeat_interleave(feature_volume_height).repeat(n_proposals).reshape(-1, 1)

        return anchors_z_cut_indices, anchors_y_indices, anchors_x_indices, invalid_mask


import logging
import os

def setup_logging(logs_dir):
    """
        Setups the logging configuration for the project.

        Returns:
            logger: The logger object.
    """

    # Create logs directory if not exists
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    # Creates the latest log file
    log_files = os.listdir(logs_dir)
    log_files = [f for f in log_files if f.endswith(".log")]
    log_files.sort()
    if len(log_files) > 0:
        number = int(log_files[-1].split("_")[1].split(".")[0])
        number += 1
        log_path = os.path.join(logs_dir, "train_{:04d}.log".format(number))
    else:
        log_path = os.path.join(logs_dir, "train_0000.log")

    # Specifies the format of the log messages and returns the logger
    formatter = logging.Formatter("[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s")
    file_handler = logging.FileHandler(log_path)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(formatter)
    logging.basicConfig(level=logging.DEBUG, handlers=[file_handler, stream_handler])

    return logging.getLogger(__name__)


from .anchors import generate_anchors, get_fv_anchor_indices
from .logger import setup_logging
from .model_state import load_last_train_state, get_last_checkpoint, save_train_state, save_data, remove_data, plot_from_data
from .dataset import LaneDataset
from .focal_loss import FocalLoss


import matplotlib.pyplot as plt
import os
import re
import torch
import pickle

def load_last_train_state(model, optimizer, scheduler, checkpoints_dir):
    """
        Load the last training state from the checkpoint files.

        Args:
            model: The model to be loaded.
            optimizer: The optimizer to be loaded.
            scheduler: The scheduler to be loaded.
            config: The configuration of the training.
        
        Returns:
            epoch: The epoch of the last checkpoint.
            model: The model loaded from the last checkpoint.
            optimizer: The optimizer loaded from the last checkpoint.
            scheduler: The scheduler loaded from the last checkpoint.
    """
    train_state_path, epoch = get_last_checkpoint(checkpoints_dir)
    train_state = torch.load(os.path.join(checkpoints_dir, train_state_path), weights_only=True)
    model.load_state_dict(train_state['model'])
    optimizer.load_state_dict(train_state['optimizer'])
    scheduler.load_state_dict(train_state['scheduler'])

    return epoch, model, optimizer, scheduler

def save_train_state(epoch, model, optimizer, scheduler, checkpoints_dir):
    """
        Save the training state to the checkpoint files.

        Args:
            epoch: The epoch of the current training state.
            model: The model to be saved.
            optimizer: The optimizer to be saved.
            scheduler: The scheduler to be saved.
    """
    if not os.path.exists(checkpoints_dir):
        os.makedirs(checkpoints_dir)
    
    train_state = {
        'epoch': epoch,
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict(),
        'scheduler': scheduler.state_dict()
    }
    torch.save(train_state, os.path.join(checkpoints_dir, f'laneatt_{epoch}.pt'))

def get_last_checkpoint(checkpoints_dir):
    """
        Get the epoch of the last checkpoint.

        Returns:
            The epoch of the last checkpoint.
    """
    if not os.path.exists(checkpoints_dir): raise FileNotFoundError('Checkpoints directory not found.')
    
    # Generate the pattern to match the checkpoint files and a list of all the checkpoint files
    pattern = re.compile('laneatt_(\\d+).pt')
    checkpoints = [ckpt for ckpt in os.listdir(checkpoints_dir) if re.match(pattern, ckpt) is not None]
    if len(checkpoints) == 0: raise FileNotFoundError('No checkpoint files found.')

    # Get last checkpoint epoch
    latest_checkpoint_path = sorted(checkpoints, reverse=True, key=lambda name : int(name.split('_')[1].rstrip('.pt')))[0]
    epoch = latest_checkpoint_path.split('_')[1].rstrip('.pt')

    return latest_checkpoint_path, int(epoch)

def save_data(data, data_dir, data_name):
    """
        Save the data to the data directory.

        Args:
            data: The data to be saved.
            data_dir: The directory to save the data.
            data_name: The name of the data file.
    """
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    
    if not os.path.exists(os.path.join(data_dir, data_name)):
        with open(os.path.join(data_dir, data_name), 'wb') as f:
            pickle.dump([data], f)
    else:
        with open(os.path.join(data_dir, data_name), 'rb') as f:
            read_data = pickle.load(f)
        read_data.append(data)
        with open(os.path.join(data_dir, data_name), 'wb') as f:
            pickle.dump(read_data, f)

def remove_data(data_dir):
    """
        Remove the data from the data directory.

        Args:
            data_dir: The directory to remove the data.
    """
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    elements = os.listdir(data_dir)
    for element in elements:
        os.remove(os.path.join(data_dir, element) if os.path.isfile(os.path.join(data_dir, element)) else None)

def plot_from_data(data_dir):
    """
        Plot the data from the data directory.

        Args:
            data_dir: The directory to plot the data.
    """
    if not os.path.exists(data_dir): raise FileNotFoundError('Data directory not found.')
    
    data_files = [f for f in os.listdir(data_dir) if f.endswith('.pkl')]

    train_loss, train_cls_loss, train_reg_loss = [], [], []
    eval_loss, eval_cls_loss, eval_reg_loss, precision, recall, f1, accuracy = [], [], [], [], [], [], []
    for data_file in data_files:
        data_type = data_file.split('_')[0]

        with open(os.path.join(data_dir, data_file), 'rb') as f:
            data = pickle.load(f)

        if data_type == 'train':
            for entry in data:
                train_loss.append(entry['loss'])
                train_cls_loss.append(entry['cls_loss'])
                train_reg_loss.append(entry['reg_loss'])
        elif data_type == 'eval':
            for entry in data:
                eval_loss.append(entry['loss'])
                eval_cls_loss.append(entry['cls_loss'])
                eval_reg_loss.append(entry['reg_loss'])
                precision.append(entry['precision'])
                recall.append(entry['recall'])
                f1.append(entry['f1_score'])
                accuracy.append(entry['accuracy'])
        
    # Plot the data
    ax, fig = plt.subplots(2, 5, figsize=(20, 10))
    fig[0, 0].plot(train_loss)
    fig[0, 0].set_title('Train Loss')
    fig[0, 1].plot(train_cls_loss)
    fig[0, 1].set_title('Train Classification Loss')
    fig[0, 2].plot(train_reg_loss)
    fig[0, 2].set_title('Train Regression Loss')
    fig[1, 0].plot(eval_loss)
    fig[1, 0].set_title('Eval Loss')
    fig[1, 1].plot(eval_cls_loss)
    fig[1, 1].set_title('Eval Classification Loss')
    fig[1, 2].plot(eval_reg_loss)
    fig[1, 2].set_title('Eval Regression Loss')
    fig[0, 3].plot(precision)
    fig[0, 3].set_title('Precision')
    fig[0, 4].plot(recall)
    fig[0, 4].set_title('Recall')
    fig[1, 3].plot(f1)
    fig[1, 3].set_title('F1 Score')
    fig[1, 4].plot(accuracy)
    fig[1, 4].set_title('Accuracy')

    plt.savefig(os.path.join(data_dir, f'metrics.png'))


from setuptools import setup

setup(
  name = 'laneatt',
  packages = ['laneatt', 'laneatt.utils'],
  package_data={'laneatt': ['config/backbones.yaml']},
  include_package_data=True,
  version = '2.4.3',
  license='MIT',
  description = 'A package to detect lane lines in images and videos',
  author = 'Paolo Reyes',
  author_email = 'paolo.alfonso.reyes@gmail.com',
  url = 'https://github.com/PaoloReyes/RealTime-LaneATT',
  download_url = 'https://github.com/PaoloReyes/RealTime-LaneATT/archive/refs/tags/LaneATT-v2.4.tar.gz',
  keywords = ['lanes', 'AI', 'greenhouse', 'detection', 'machine learning', 'laneatt', 'delimitations'],
  install_requires=[
          'filelock',
          'fsspec',
          'Jinja2',
          'joblib',
          'MarkupSafe',
          'mpmath',
          'matplotlib',
          'networkx',
          'numpy',
          'opencv-python',
          'pillow',
          'PyYAML',
          'scikit-learn',
          'scipy',
          'setuptools',
          'sympy',
          'threadpoolctl',
          'torch',
          'torchaudio',
          'torchvision',
          'tqdm',
          'typing_extensions',
          'wheel',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable', # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package

    'Intended Audience :: Developers', # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',

    'License :: OSI Approved :: MIT License', # License

    'Programming Language :: Python :: 3', # Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.12',
  ],
)


from . import utils
from torchvision import models
from torchvision.transforms import ToTensor
from tqdm import tqdm, trange

import cv2
import os 
import random
import torch
import yaml

import numpy as np
import torch.nn as nn

class LaneATT(nn.Module):
    def __init__(self, config:str) -> None:
        """
            LaneATT model initialization.

            Args:
                config (str): Path to the configuration file
        """
        super(LaneATT, self).__init__()

        # Config file
        self.__laneatt_config = yaml.safe_load(open(config))

        # Load backbones config file
        self.__backbones_config = yaml.safe_load(open(os.path.join(os.path.dirname(__file__), 'config', 'backbones.yaml')))

        # Set anchor feature channels
        self.__feature_volume_channels = self.__laneatt_config['feature_volume_channels']

        # Set anchor y discretization
        self.__anchor_y_discretization = self.__laneatt_config['anchor_discretization']['y']

        # Set anchor x steps
        self.__anchor_x_discretization = self.__laneatt_config['anchor_discretization']['x']

        # Set image width and height
        self.__img_w = self.__laneatt_config['image_size']['width']
        self.__img_h = self.__laneatt_config['image_size']['height']

        # Create anchor feature dimensions variables but they will be defined after the backbone is created
        self.__feature_volume_height = None
        self.__feature_volume_width = None

        # Set device
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        # Creates the backbone and moves it to the device
        self.backbone = self.__laneatt_config['backbone']

        # Positive Threshold
        self.__positive_threshold = self.__laneatt_config['positive_threshold']

        # Generate Anchors Proposals
        self.__anchors_image, self.__anchors_feature_volume = utils.generate_anchors(y_discretization=self.__anchor_y_discretization, 
                                                                                    x_discretization=self.__anchor_x_discretization,
                                                                                    left_angles=self.__laneatt_config['anchor_angles']['left'],
                                                                                    right_angles=self.__laneatt_config['anchor_angles']['right'],
                                                                                    bottom_angles=self.__laneatt_config['anchor_angles']['bottom'],
                                                                                    fv_size=(self.__feature_volume_channels, 
                                                                                             self.__feature_volume_height, 
                                                                                             self.__feature_volume_width),
                                                                                    img_size=(self.__img_h, self.__img_w))
        
        # Move the anchors to the device
        self.__anchors_image = self.__anchors_image.to(self.device)
        self.__anchors_feature_volume = self.__anchors_feature_volume.to(self.device)

        # Pre-Compute Indices for the Anchor Pooling
        self.__anchors_z_indices, self.__anchors_y_indices, self.__anchors_x_indices, self.__invalid_mask = utils.get_fv_anchor_indices(self.__anchors_feature_volume,
                                                                                                                                        self.__feature_volume_channels, 
                                                                                                                                        self.__feature_volume_height, 
                                                                                                                                        self.__feature_volume_width)

        # Move the indices to the device
        self.__anchors_z_indices = self.__anchors_z_indices.to(self.device)
        self.__anchors_y_indices = self.__anchors_y_indices.to(self.device)
        self.__anchors_x_indices = self.__anchors_x_indices.to(self.device)
        self.__invalid_mask = self.__invalid_mask.to(self.device)

        # Fully connected layer of the attention mechanism that takes a single anchor proposal for all the feature maps as input and outputs a score 
        # for each anchor proposal except itself. The score is computed using a softmax function.
        self.__attention_layer = nn.Sequential(nn.Linear(self.__feature_volume_channels * self.__feature_volume_height, len(self.__anchors_feature_volume) - 1),
                                                nn.Softmax(dim=1)).to(self.device)
        
        # Convolutional layer for the classification and regression tasks
        self.__cls_layer = nn.Linear(2 * self.__feature_volume_channels * self.__feature_volume_height, 2).to(self.device)
        self.__reg_layer = nn.Linear(2 * self.__feature_volume_channels * self.__feature_volume_height, self.__anchor_y_discretization + 1).to(self.device)

        self.__focal_loss = utils.FocalLoss(alpha=0.25, gamma=2.)

    @property
    def backbone(self) -> nn.Module:
        """
            Get the backbone (RESNET) of the model
        """
        return self.__backbone
    
    @backbone.setter
    def backbone(self, value:str) -> None:
        """
            Set the backbone for the model taking into account available backbones in the config file
            It cuts the average pooling and fully connected layer from the backbone and adds a convolutional 
            layer to reduce the dimensionality to the desired feature volume channels and moves the model 
            to the device

            Args:
                value (str): Backbone name
            
            Raises:
                ValueError: If the backbone is not in the list of backbones in the config file
        """
        # Lower the value to avoid case sensitivity
        value = value.lower()

        # Check if value is in the list of backbones in config file
        if value not in self.__backbones_config['backbones']:
            raise ValueError(f'Backbone must be one of {self.config["backbones"]}')
        
        # Set pretrained backbone according to pytorch requirements without the average pooling and fully connected layer
        self.__backbone = nn.Sequential(*list(models.__dict__[value](weights=f'{value.replace("resnet", "ResNet")}_Weights.DEFAULT').children())[:-2],)

        # Runs backbone (on cpu) once to get output data 
        backbone_dimensions = self.__backbone(torch.randn(1, 3, self.__img_h, self.__img_w)).shape

        # Extracts feature volume height and width
        self.__feature_volume_height = backbone_dimensions[2]
        self.__feature_volume_width = backbone_dimensions[3]

        # Join the backbone and the convolutional layer for dimensionality reduction
        self.__backbone = nn.Sequential(self.__backbone, nn.Conv2d(backbone_dimensions[1], self.__feature_volume_channels, kernel_size=1))

        # Move the model to the device
        self.__backbone.to(self.device)

    @property
    def img_w(self) -> int:
        return self.__img_w
    
    @property
    def img_h(self) -> int:
        return self.__img_h

    def forward(self, x:torch.Tensor) -> torch.Tensor:
        """
            Forward pass of the model

            Args:
                x (torch.Tensor): Input image

            Returns:
                torch.Tensor: Regression proposals
        """
        # Move the input to the device
        x = x.to(self.device)
        # Gets the feature volume from the backbone with a dimensionality reduction layer
        feature_volumes = self.backbone(x)
        # Extracts the anchor features from the feature volumes
        batch_anchor_features = self.__cut_anchor_features(feature_volumes)
        # Join proposals from all feature volume channels into a single dimension and stacks all the batches
        batch_anchor_features = batch_anchor_features.view(-1, self.__feature_volume_channels * self.__feature_volume_height)

        # Compute attention scores and reshape them to the original batch size
        attention_scores = self.__attention_layer(batch_anchor_features).reshape(x.shape[0], len(self.__anchors_feature_volume), -1)
        # Generate the attention matrix to be used to store the attention scores
        attention_matrix = torch.eye(attention_scores.shape[1], device=x.device).repeat(x.shape[0], 1, 1)
        # Gets the indices of the non diagonal elements of the attention matrix
        non_diag_indices = torch.nonzero(attention_matrix == 0., as_tuple=False)
        # Makes the entire attention matrix to be zero
        attention_matrix[:] = 0
        # Assigns the attention scores to the attention matrix ignoring the self attention scores as they are not calculated
        # This way we can have a matrix with the attention scores for each anchor proposal
        attention_matrix[non_diag_indices[:, 0], non_diag_indices[:, 1], non_diag_indices[:, 2]] = attention_scores.flatten()

        # Reshape the batch anchor features to the original batch size
        batch_anchor_features = batch_anchor_features.reshape(x.shape[0], len(self.__anchors_feature_volume), -1)
        # Computes the attention features by multiplying the anchor features with the attention weights per batch
        # This will give more context based on the probability of the current anchor to be a lane line compared to other frequently co-occurring anchor proposals
        # And adds them into a single tensor implicitly by using a matrix multiplication
        attention_features = torch.bmm(torch.transpose(batch_anchor_features, 1, 2),
                                       torch.transpose(attention_matrix, 1, 2)).transpose(1, 2)

        # Reshape the attention features batches to one batch size
        attention_features = attention_features.reshape(-1, self.__feature_volume_channels * self.__feature_volume_height)
        # Reshape the batch anchor features batches to one batch size
        batch_anchor_features = batch_anchor_features.reshape(-1, self.__feature_volume_channels * self.__feature_volume_height)

        # Concatenate the attention features with the anchor features
        batch_anchor_features = torch.cat((attention_features, batch_anchor_features), dim=1)

        # Predict the class of the anchor proposals
        cls_logits = self.__cls_layer(batch_anchor_features)
        # Predict the regression of the anchor proposals
        reg = self.__reg_layer(batch_anchor_features)

        # Undo joining the proposals from all images into proposals features batches
        cls_logits = cls_logits.reshape(x.shape[0], -1, cls_logits.shape[1])
        reg = reg.reshape(x.shape[0], -1, reg.shape[1])
        
        # Create the regression proposals
        reg_proposals = torch.zeros((*cls_logits.shape[:2], 5 + self.__anchor_y_discretization), device=self.device)
        # Assign the anchor proposals to the regression proposals
        reg_proposals += self.__anchors_image
        # Assign the classification scores to the regression proposals
        reg_proposals[:, :, :2] = cls_logits
        # Adds the regression offsets to the anchor proposals in the regression proposals
        reg_proposals[:, :, 4:] += reg

        return reg_proposals
    
    def cv2_inference(self, frame:np.ndarray) -> torch.Tensor:
        """
            Inference of the model using OpenCV frame

            Args:
                image (np.ndarray): Image
            
            Returns:
                torch.Tensor: Regression proposals
        """
        frame = cv2.resize(frame, (self.img_w, self.img_h))
        img_tensor = ToTensor()((frame.copy()/255.0).astype(np.float32)).permute(0, 1, 2)
        output = self.forward(img_tensor.unsqueeze(0)).squeeze(0)

        return self.postprocess(output)

    def postprocess(self, output:torch.Tensor) -> torch.Tensor:
        """
            Postprocess the regression proposals

            Args:
                output (torch.Tensor): Regression proposals

            Returns:
                torch.Tensor: Good proposals
        """
        # Filter proposals with confidence below the threshold
        good_proposals = output[output[:, 1] > self.__positive_threshold]
        return good_proposals
    
    def nms(self, output:torch.Tensor, nms_threshold:float=40.0) -> torch.Tensor:
        """
            Apply non-maximum suppression to the proposals

            Args:
                output (torch.Tensor): Regression proposals
                threshold (float): Threshold for valid proposals
                nms_threshold (float): NMS threshold

            Returns:
                torch.Tensor: Good proposals NMS suppressed 
        """
        # Filter proposals with confidence below the threshold and sort them by confidence
        good_proposals = output[output[:, 1] > self.__positive_threshold]
        good_proposals = good_proposals[good_proposals[:, 3].argsort(descending=True)]
        # Verify if there are no proposals
        if len(good_proposals) == 0: return good_proposals

        # Create a mask to store the same line proposals
        good_proposals_mask = np.zeros((len(good_proposals), len(good_proposals)), dtype=bool)
        # Iterate over the proposals
        for i, line_a in enumerate(good_proposals):
            # Get the start and end of the current proposal
            start_a = line_a[2] / self.__img_h * self.__anchor_y_discretization
            end_a = line_a[2] + line_a[4]
            # Iterate over the rest of the proposals
            for j, line_b in enumerate(good_proposals):
                # Get the start and end of the current comparison proposal
                start_b = line_b[2] / self.__img_h * self.__anchor_y_discretization
                end_b = line_b[2] + line_b[4] - 1
                
                # Get the start and end intersection between the proposals
                start = int(max(start_a, start_b))
                end = int(min(end_a, end_b, self.__anchor_y_discretization))

                # Calculate the error between the proposals
                error = torch.tensor(0., device=self.device)
                for k in range(start, end):
                    error += abs(line_a[5 + k] - line_b[5 + k])
                error /= end - start
                error = error.item()

                # If the error is below the NMS threshold, we consider the proposals to represent the same line
                good_proposals_mask[i][j] = error < nms_threshold

        # List to store the indexes of the unique lines
        unique_line_indexes = [0]
        while True:
            # Get a unique line
            line = good_proposals_mask[unique_line_indexes[-1]]
            found_different = False
            # Iterate over a unique line against the rest of the proposals errors
            for i, cmp_line in enumerate(line):
                # If the line is different and the index is greater than the last unique line index we found a different line
                # so we append it to the unique line indexes
                if not cmp_line and i > unique_line_indexes[-1]:
                    unique_line_indexes.append(i)
                    found_different = True
                    break
            
            # If we stop finding different lines, we break the loop
            if not found_different:
                break

        # Based on the unique line indexes, we get a range of similar lines and get the one with the highest confidence
        # Create a list to store the high confidence unique line indexes
        high_confidence_unique_line_indexes = [0 for _ in range(len(unique_line_indexes))]
        # Iterate over the unique line indexes
        for i in range(len(unique_line_indexes)):
            # Verify if we are in the last unique line index
            if i == len(unique_line_indexes) - 1:
                # If so, we get the highest confidence line from the last unique line index to the end
                high_confidence_unique_line_indexes[i] = good_proposals[unique_line_indexes[i]:][:, 1].argmax().item()
            else:
                # Otherwise, we get the highest confidence line from the current unique line index to the next unique line index
                high_confidence_unique_line_indexes[i] = good_proposals[unique_line_indexes[i]:unique_line_indexes[i+1]][:, 1].argmax().item()
            
            # Add an offset to counteract for the list slicing
            high_confidence_unique_line_indexes[i] += unique_line_indexes[i]
                
        return good_proposals[unique_line_indexes]
                
    def plot(self, output:torch.Tensor, image:np.ndarray) -> None:
        """
            Plot the lane lines on the image

            Args:
                output (torch.Tensor): Regression proposals
                image (np.ndarray): Image
        """
        proposals_length = output[:, 4]
        # Get the y discretization values
        ys = torch.linspace(self.__img_h, 0, self.__anchor_y_discretization, device=self.device)
        # Store x and y values for each lane line
        output = [[(x, ys[i]) for i, x in enumerate(lane[5:])] for lane in output]

        # Resize the image to the model's trained size
        img = cv2.resize(image, (self.__img_w, self.__img_h))
        # Iterate over the lanes
        for i, lane in enumerate(output):
            # Internal loop variables to account for the first point and the change in color of the lines
            prev_x, prev_y = lane[0]
            color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
            # Iterate over the line points
            for j, (x, y) in enumerate(lane):
                # Break the loop if the proposal length is reached
                if int(proposals_length[i].item()) == j: break
                # Draw a line between the previous point and the current point
                cv2.line(img, (int(prev_x), int(prev_y)), (int(x), int(y)), color, 2)
                prev_x, prev_y = x, y

        # Show the image
        cv2.imshow('frame', img)

    def train_model(self, resume:bool=False):
        """
            Train the model
        """
        if not resume:
            utils.remove_data(self.__laneatt_config['outputs_dir'])
        # Setup the logger
        self.logger = utils.setup_logging(self.__laneatt_config['logs_dir'])

        self.logger.info('Starting training...')

        model = self.to(self.device)

        # Get the optimizer and the scheduler from the config file
        optimizer = getattr(torch.optim, self.__laneatt_config['optimizer']['name'])(model.parameters(), **self.__laneatt_config['optimizer']['parameters'])
        scheduler = getattr(torch.optim.lr_scheduler, self.__laneatt_config['lr_scheduler']['name'])(optimizer, **self.__laneatt_config['lr_scheduler']['parameters'])

        # State the starting epoch
        starting_epoch = 1
        # Load the last training state if the resume flag is set and modify the starting epoch and model
        if resume:
            last_epoch, model, optimizer, scheduler = utils.load_last_train_state(model, optimizer, scheduler, self.__laneatt_config['checkpoints_dir'])
            starting_epoch = last_epoch + 1
        
        # Get the number of epochs from the config file and the train loader
        epochs = self.__laneatt_config['epochs']
        train_loader = self.__get_dataloader('train')

        # Iterate over the epochs
        for epoch in trange(starting_epoch, epochs + 1, initial=starting_epoch - 1, total=epochs):
            self.logger.info('Epoch [%d/%d] starting.', epoch, epochs)
            # Put the model in training mode
            model.train()
            # Get the progress bar object based on the train loader
            pbar = tqdm(train_loader)
            # Iterate over the batches
            accumulator = {'cls_loss': 0, 'reg_loss': 0, 'loss': 0}
            for i, (images, labels) in enumerate(pbar):
                # Move the images and labels to the device
                images = images.to(self.device)
                labels = labels.to(self.device)
                
                # Forward pass
                outputs = model(images)
                loss, loss_dict_i = model.__loss(outputs, labels)

                # Backward and optimize
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                # Scheduler step (iteration based)
                scheduler.step()

                # Log
                postfix_dict = {key: float(value) for key, value in loss_dict_i.items()}
                postfix_dict['lr'] = optimizer.param_groups[0]["lr"]

                accumulator['cls_loss'] += loss_dict_i['cls_loss'].to('cpu').item()
                accumulator['reg_loss'] += loss_dict_i['reg_loss'].to('cpu').item()
                accumulator['loss'] += loss.to('cpu').item()

                line = 'Epoch [{}/{}] - Iter [{}/{}] - Loss: {:.5f} - '.format(epoch, epochs, i, len(train_loader), loss.item())
                line += ' - '.join(['{}: {:.5f}'.format(component, postfix_dict[component]) for component in postfix_dict])
                self.logger.debug(line)
                
                postfix_dict['loss'] = loss.item()
                pbar.set_postfix(ordered_dict=postfix_dict)
            
            # Compute the average loss
            for key in accumulator:
                accumulator[key] /= len(train_loader)

            self.logger.debug('Epoch [%d/%d] finished.', epoch, epochs)
            self.logger.info('Training Epoch finished. Loss: {:.5f} - Cls Loss: {:.5f} - Reg Loss: {:.5f}'.format(accumulator['loss'], accumulator['cls_loss'], accumulator['reg_loss']))

            # Save the data in a pickle file
            utils.save_data(accumulator, self.__laneatt_config['outputs_dir'], 'train_data.pkl')

            self.eval_model()

            utils.plot_from_data(self.__laneatt_config['outputs_dir'])
            # Save the model state every model_checkpoint_interval epochs from the config file
            if epoch % self.__laneatt_config['model_checkpoint_interval'] == 0:
                utils.save_train_state(epoch, model, optimizer, scheduler, self.__laneatt_config['checkpoints_dir'])

    def eval_model(self):
        """
            Evaluate the model
        """
        self.logger.info('Starting evaluation...')

        model = self.to(self.device)

        # Get the test loader
        val_loader = self.__get_dataloader('val')

        # Put the model in evaluation mode
        model.eval()
        # Get the progress bar object based on the test loader
        pbar = tqdm(val_loader)
        # Iterate over the batches
        accumulator = {'cls_loss': 0, 'reg_loss': 0, 'loss': 0}
        precision, recall, accuracy = 0, 0, 0
        with torch.no_grad():
            for i, (images, labels) in enumerate(pbar):
                # Move the images and labels to the device
                images = images.to(self.device)
                labels = labels.to(self.device)
                
                # Forward pass
                outputs = model(images)
                loss, loss_dict_i = model.__loss(outputs, labels)

                p, r, acc = self.__presicion_recall_accuracy(outputs, labels)
                precision += p
                recall += r
                accuracy += acc

                accumulator['cls_loss'] += loss_dict_i['cls_loss'].to('cpu').item()
                accumulator['reg_loss'] += loss_dict_i['reg_loss'].to('cpu').item()
                accumulator['loss'] += loss.to('cpu').item()

                postfix_dict = {key: float(value) for key, value in loss_dict_i.items()}
                line = 'Iter [{}/{}] - Loss: {:.5f} - '.format(i, len(val_loader), loss.item())
                line += ' - '.join(['{}: {:.5f}'.format(component, postfix_dict[component]) for component in postfix_dict])
                self.logger.debug(line)

                postfix_dict['loss'] = loss.item()
                pbar.set_postfix(ordered_dict=postfix_dict)
        
        # Compute the average loss
        for key in accumulator:
            accumulator[key] /= len(val_loader)
        
        precision /= len(val_loader)
        recall /= len(val_loader)
        accuracy /= len(val_loader)
        f1_score = 2 * (precision * recall) / (precision + recall + 1e-6)

        output_dir = accumulator.copy()
        output_dir['precision'] = precision
        output_dir['recall'] = recall
        output_dir['f1_score'] = f1_score
        output_dir['accuracy'] = accuracy

        utils.save_data(output_dir, self.__laneatt_config['outputs_dir'], 'eval_data.pkl')
        
        self.logger.info('Evaluation finished. Loss: {:.5f} - Cls Loss: {:.5f} - Reg Loss: {:.5f} - Precision: {:.5f} - Recall: {:.5f} - F1: {:.5f} - Acc: {:.5f}'.format(accumulator['loss'], accumulator['cls_loss'], accumulator['reg_loss'], precision, recall, f1_score, accuracy))

    def __presicion_recall_accuracy(self, proposals_list:torch.Tensor, targets:torch.Tensor) -> torch.Tensor:
        """
            Compute the precision and recall of the model

            Args:
                proposals_list (torch.Tensor): Tensor of proposals
                targets (torch.Tensor): Tensor of targets
                threshold (float): Threshold for valid proposals

            Returns:
                torch.Tensor: Precision and Recall values
        """
        true_positives, false_positives, false_negatives, accuracy = 0, 0, 0, 0
        for proposals, target in zip(proposals_list, targets):
            target = target[target[:, 1] == 1]
            # The model outputs a classification score and a regression offset for each anchor proposal
            # So we select anchors that are most similar to the ground truth lane lines using the positive mask
            positives = proposals[proposals[:, 1] > self.__positive_threshold]
            line_predictions = self.nms(positives)

            if len(line_predictions) == 0:
                tp, fp = 0, 0
                fn = target.shape[0]
                acc = fn == 0
            else:
                _, _, _, _, tp, fp, fn, acc = self.__match_proposals_with_targets(line_predictions, target, metrics=True)
                tp, fp, fn, acc = int(tp.item()), int(fp.item()), int(fn.item()), int(acc.item())

            true_positives += tp
            false_positives += fp
            false_negatives += fn
            accuracy += acc
        
        precision = true_positives / (true_positives + false_positives + 1e-6)
        recall = true_positives / (true_positives + false_negatives + 1e-6)
        accuracy /= len(proposals_list)

        return precision, recall, accuracy

    def load(self, checkpoint:str):
        """
            Load the model from a checkpoint file

            Args:
                checkpoint (str): Checkpoint file
        """
        train_state = torch.load(checkpoint, weights_only=True)
        self.load_state_dict(train_state['model'])

    def __cut_anchor_features(self, feature_volumes:torch.Tensor) -> torch.Tensor:
        """
            Extracts anchor features from the feature volumes

            Args:
                feature_volumes (torch.Tensor): Feature volumes

            Returns:
                torch.Tensor: Anchor features (n_proposals, n_channels, n_height, 1)
        """

        # Gets the batch size
        batch_size = feature_volumes.shape[0]
        # Gets the number of anchor proposals
        anchor_proposals = len(self.__anchors_feature_volume)
        # Builds a tensor to store the anchor features 
        batch_anchor_features = torch.zeros((batch_size, anchor_proposals, self.__feature_volume_channels, self.__feature_volume_height, 1), 
                                            device=self.device)
        
        # Iterates over each batch
        for batch_idx, feature_volume in enumerate(feature_volumes):
            # Extracts features from the feature volume using the anchor indices, the output will be in a single dimension
            # so we reshape it to a new volume with proposals in the channel dimension, fv_channels in the width dimension
            # and fv_height in the height dimension. So the features extracted from each feature map for each proposal
            # will be in the same channel storing the features of the anchor proposals in each proposed index in the height dimension
            rois = feature_volume[self.__anchors_z_indices, 
                                  self.__anchors_y_indices, 
                                  self.__anchors_x_indices].view(anchor_proposals, self.__feature_volume_channels, self.__feature_volume_height, 1)
            
            # Sets to zero the anchor proposals that are outside the feature map to avoid taking the edge values
            rois[self.__invalid_mask] = 0
            # Assigns the anchor features to the batch anchor features tensor
            batch_anchor_features[batch_idx] = rois

        return batch_anchor_features

    def __loss(self, proposals_list:torch.Tensor, targets:torch.Tensor, cls_loss_weight:int=10) -> torch.Tensor:
        """
            Compute the loss of the model

            Args:
                proposals_list (torch.Tensor): Tensor of proposals
                targets (torch.Tensor): Tensor of targets

            Returns:
                torch.Tensor: Loss value in Tensor format
        """
        # Initialize the variables for the losses
        smooth_l1_loss = nn.SmoothL1Loss()
        cls_loss = 0
        reg_loss = 0
        valid_imgs = len(targets)
        total_positives = 0
        # Iterate over each batch
        for proposals, target in zip(proposals_list, targets):
            # Filter lane targets that do not exist (confidence == 0)
            target = target[target[:, 1] == 1]

            # Match proposals with targets to get useful indices
            with torch.no_grad():
                positives_mask, invalid_offsets_mask, negatives_mask, target_positives_indices, _, _, _, _ = self.__match_proposals_with_targets(self.__anchors_image, target)

            # The model outputs a classification score and a regression offset for each anchor proposal
            # So we select anchors that are most similar to the ground truth lane lines using the positive mask
            positives = proposals[positives_mask]

            num_positives = len(positives)
            total_positives += num_positives
            # Select anchors that are not similar to the ground truth lane lines using the negative mask
            negatives = proposals[negatives_mask]
            num_negatives = len(negatives)

            # Handle edge case of no positives found by setting targets to 0 and comparing to the classification scores for all proposals that should be 0
            # in a perfect model
            if num_positives == 0:
                cls_target = proposals.new_zeros(len(proposals)).long()
                cls_pred = proposals[:, :2]
                cls_loss += self.__focal_loss(cls_pred, cls_target).sum()
                continue

            # Concatenate positives and negatives
            all_proposals = torch.cat((positives, negatives), dim=0)
            # Create a tensor containing 1 for positives and 0 for negatives
            cls_target = proposals.new_zeros(num_positives + num_negatives).long()
            cls_target[:num_positives] = 1.
            # Get the classification scores
            cls_pred = all_proposals[:, :2]

            # Regression from the positive anchors
            reg_pred = positives[:, 4:]
            with torch.no_grad():
                # Extract the targets that are matched with the positive anchors, could be repeated
                target = target[target_positives_indices]
                # Get the start index of the positive anchors and the target
                positive_starts = (positives[:, 2] / self.__img_h * self.__anchor_y_discretization).round().long()
                target_starts = (target[:, 2]  / self.__img_h * self.__anchor_y_discretization).round().long()
                # Adjust the target length according to the start intersection
                target[:, 4] -= positive_starts - target_starts
                # Create a tensor to store an index for each model output
                all_indices = torch.arange(num_positives, device=self.device, dtype=torch.long)
                # Get the end index of the intersection based on the positive start index and the target length
                ends = (positive_starts + target[:, 4] - 1).round().long()
                # Uses the same trick as matching proposals with targets to get the invalid offsets mask
                invalid_offsets_mask = torch.zeros((num_positives, 1 + self.__anchor_y_discretization + 1), device=self.device, dtype=torch.int)  # length + S + pad
                invalid_offsets_mask[all_indices, 1 + positive_starts] = 1
                invalid_offsets_mask[all_indices, 1 + ends + 1] -= 1
                invalid_offsets_mask = invalid_offsets_mask.cumsum(dim=1) == 0
                invalid_offsets_mask = invalid_offsets_mask[:, :-1]
                invalid_offsets_mask[:, 0] = False
                # Get the regression target
                reg_target = target[:, 4:]
                # Get the regression prediction where no intersection is found and assign it to the target
                # This is done to complete incomplete targets and do not consider them in the regression loss
                # Because even though targets are not full height, predictions are made for the full height
                # And we have to counteract this effect
                reg_target[invalid_offsets_mask] = reg_pred[invalid_offsets_mask]

            # # Loss calc
            reg_loss += smooth_l1_loss(reg_pred, reg_target)
            # Get the classification loss per anchor, adds them to get entire loss and divide by the number of positives 
            # It is not very clear why divide by the number of positives, but it might be to compensate for the number of
            # positives in the batch
            cls_loss += self.__focal_loss(cls_pred, cls_target).sum() / num_positives

        # Batch mean
        cls_loss /= valid_imgs
        reg_loss /= valid_imgs

        # Total loss
        loss = cls_loss_weight * cls_loss + reg_loss
        return loss, {'cls_loss': cls_loss, 'reg_loss': reg_loss, 'batch_positives': total_positives}

    def __match_proposals_with_targets(self, proposals:torch.Tensor, targets:torch.Tensor, t_pos:int=15., t_neg:int=20., metrics=False) -> torch.Tensor:
        """
            Match anchor proposals with targets

            Args:
                proposals (torch.Tensor): Anchor proposals projected to the image space
                targets (torch.Tensor): Ground truth lane lines
                t_pos (float): Positive threshold
                t_neg (float): Negative threshold

            Returns:
                torch.Tensor(num_anchor_proposals): A boolean tensor indicating if the anchor proposal is a positive
                torch.Tensor(num_positives, y_discretization+5): A boolean tensor indicating offsets that do not intersect with the target
                torch.Tensor(num_anchor_proposals): A boolean tensor indicating if the anchor proposal is a negative
                torch.Tensor(num_positives): A tensor with the indices of the target matched with the positive anchor proposal
        """
        num_proposals = proposals.shape[0]
        num_targets = targets.shape[0]
        # Pad proposals and target for the valid_offset_mask's trick
        proposals_pad = proposals.new_zeros(proposals.shape[0], proposals.shape[1] + 1)
        proposals_pad[:, :-1] = proposals
        proposals = proposals_pad
        targets_pad = targets.new_zeros(targets.shape[0], targets.shape[1] + 1)
        targets_pad[:, :-1] = targets
        targets = targets_pad
        num_targets = targets.shape[0]

        # Repeat targets and proposals to compare all combinations
        proposals = torch.repeat_interleave(proposals, num_targets, dim=0)
        targets = torch.cat(num_proposals * [targets])

        # Get start index of the proposals and targets
        targets_starts = targets[:, 2] / self.__img_h * self.__anchor_y_discretization
        proposals_starts = proposals[:, 2] / self.__img_h * self.__anchor_y_discretization
        # Get the start index for the intersection
        starts = torch.max(targets_starts, proposals_starts).round().long()
        # Get the end index for the target line
        ends = (targets_starts + targets[:, 4] - 1).round().long()
        # Calculate the length of the intersection
        lengths = ends - starts + 1
        # In the edge case where the intersection is negative, we set the start to the end to achieve the valid_offset_mask's trick
        ends[lengths < 0] = starts[lengths < 0] - 1
        # Since we modify the ends, we need to recalculate the lengths
        lengths[lengths < 0] = 0

        # Generate the valid_offsets_mask that will contain the valid intersection between the proposals and the targets for all combinations
        valid_offsets_mask = targets.new_zeros(targets.shape)
        all_indices = torch.arange(valid_offsets_mask.shape[0], dtype=torch.long, device=targets.device)
        # Put a one on the `start` index and a -1 on the `end` index
        # The -1 is subtracted to account for the case where the length is zero
        valid_offsets_mask[all_indices, 5 + starts] = 1.
        valid_offsets_mask[all_indices, 5 + ends + 1] -= 1.
        # Cumsum to get the valid offsets
        # Valid offsets mask before [0, 0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 0]
        # Valid offsets mask after [0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0]
        # And parse it to a boolean mask [False, False, False, True, True, True, True, False, False, False, False, False]
        valid_offsets_mask = valid_offsets_mask.cumsum(dim=1) != 0.
        # Get the invalid offsets mask by inverting the valid offsets mask
        invalid_offsets_mask = ~valid_offsets_mask

        # Compute distances between proposals and targets only inside the intersection
        # Proposals and targets errors
        errors = (targets - proposals)
        # Get only the errors that are inside the intersection
        errors = errors * valid_offsets_mask.float()
        # Get the average distance between the proposals and the targets
        distances = torch.abs(errors).sum(dim=1) / (lengths.float() + 1e-9) # Avoid division by zero

        # For those distances where the length is zero, we set the distance to a very high number since we do not want to consider them
        distances[lengths == 0] = 987654.
        # Reshape the invalid offsets mask to separate the proposals and the targets, so we the invalid mask 
        # of all targets compared to all proposals. And can be indexed by [proposal_idx, target_idx]
        invalid_offsets_mask = invalid_offsets_mask.view(num_proposals, num_targets, -1)
        # Reshape the distances to separate the proposals and the targets, so we can index the distances by [proposal_idx, target_idx]
        errors = errors.view(num_proposals, num_targets, -1)
        distances = distances.view(num_proposals, num_targets)  # d[i,j] = distance from proposal i to target j

        # Get the positives and negatives based on the distances
        # This means for each proposal, we get the target with the minimum distance average error and check if it is below the positive threshold,
        # if it is, then it is a positive, otherwise we check if it is above the negative threshold, if it is, then it is a negative.
        # There is a hysteresis between the positive and negative thresholds to avoid uncertain predictions to pass as either positive or negative
        positives = distances.min(dim=1)[0] < t_pos
        negatives = distances.min(dim=1)[0] > t_neg

        # Verify if there are positives
        if positives.sum() == 0:
            # If there are no positives, we set the target positives indices to an empty tensor
            target_positives_indices = torch.tensor([], device=positives.device, dtype=torch.long)
        else:
            # If there are positives, we get the target positives indices by
            # selecting from distances only the proposals with at least one positive (that should be in the positives tensor which is a boolean mask)
            # and get the index of the minimum distance for each proposal.
            target_positives_indices = distances[positives].argmin(dim=1)

        # Finally we update invalid_offsets_mask to only consider the masks of targets that have been matched with a positive proposal
        invalid_offsets_mask = invalid_offsets_mask[positives, target_positives_indices]

        true_positives, false_positives, false_negatives, accuracy = 0, 0, 0, 0
        if metrics:
            all_indices = torch.arange(num_proposals, device=distances.device, dtype=torch.long)
            errors = errors[all_indices, distances.argmin(dim=1)]
            errors = errors[:, :-1]
            accurate_errors = errors < 20
            accurate_errors = accurate_errors.sum(dim=1) / accurate_errors.shape[1]

            targets_covered = torch.zeros(distances.shape[1])
            proposals_covered = torch.zeros(distances.shape[0])
            for i in range(distances.shape[0]):
                for j in range(distances.shape[1]):
                    if distances[i, j] < t_pos:
                        targets_covered[j] = 1
                        proposals_covered[i] = 1

            true_positives = targets_covered.sum()
            false_positives = (proposals_covered == 0).sum()
            false_negatives = (targets_covered == 0).sum()
            accuracy = accurate_errors.sum() / accurate_errors.shape[0]

        return positives, invalid_offsets_mask[:, :-1], negatives, target_positives_indices, true_positives, false_positives, false_negatives, accuracy

    def __get_dataloader(self, split:str) -> torch.utils.data.DataLoader:
        """
            Get the dataloader object

            Args:
                split (str): Split

            Returns:
                torch.utils.data.DataLoader: Dataloader object
        """
        # Create the dataset object based on TuSimple architecture
        train_dataset = utils.LaneDataset(self.__laneatt_config, split)
        # Create the dataloader object
        train_loader = torch.utils.data.DataLoader(dataset=train_dataset,
                                                   batch_size=self.__laneatt_config['batch_size'],
                                                   shuffle=True,
                                                   worker_init_fn=self.__worker_init_fn_)
        return train_loader
    
    @staticmethod
    def __worker_init_fn_(_):
        """
            Worker initialization function

            Args:
                _ (int): Worker id

            Notes:
                This function is used to set the seed of the workers to avoid randomness
        """
        torch_seed = torch.initial_seed()
        np_seed = torch_seed // 2**32 - 1
        random.seed(torch_seed)
        np.random.seed(np_seed)